# Copyright 2016, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
from collections import namedtuple

from .descriptors import DiffResultDescriptor
from .difference import Difference


modified_properties_record = namedtuple("modified_property", ["property", "added", "removed"])


class PropertiesDifference(Difference):

    """
    Determine the difference in policy properties
    (unknown permissions, MLS, etc.) between two policies.
    """

    modified_properties = DiffResultDescriptor("diff_properties")

    def diff_properties(self):
        self.modified_properties = []

        if self.left_policy.handle_unknown != self.right_policy.handle_unknown:
            self.modified_properties.append(
                modified_properties_record("handle_unknown",
                                           self.right_policy.handle_unknown,
                                           self.left_policy.handle_unknown))

        if self.left_policy.mls != self.right_policy.mls:
            self.modified_properties.append(
                modified_properties_record("MLS",
                                           self.right_policy.mls,
                                           self.left_policy.mls))

        if self.left_policy.version != self.right_policy.version:
            self.modified_properties.append(
                modified_properties_record("version",
                                           self.right_policy.version,
                                           self.left_policy.version))

    #
    # Internal functions
    #
    def _reset_diff(self):
        """Reset diff results on policy changes."""
        self.log.debug("Resetting property differences")
        self.modified_properties = None



# Copyright 2014-2015, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SETools.  If not, see <http://www.gnu.org/licenses/>.
#
import os
import unittest

from setools import InfoFlowAnalysis
from setools import TERuletype as TERT
from setools.exception import InvalidType
from setools.permmap import PermissionMap
from setools.policyrep import Type

from . import mixins
from .policyrep.util import compile_policy


# Note: the testing for having correct rules on every edge is only
# performed once on the full graph, since it is assumed that NetworkX's
# Digraph.subgraph() function correctly copies the edge attributes into
# the subgraph.


class ConditionalInfoFlowAnalysisTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.p = compile_policy("tests/conditionalinfoflow.conf", mls=False)
        cls.m = PermissionMap("tests/perm_map")
        cls.a = InfoFlowAnalysis(cls.p, cls.m)

    @classmethod
    def tearDownClass(cls):
        os.unlink(cls.p.path)

    def test_001_keep_conditional_rules(self):
        """Keep all conditional rules."""
        self.a.booleans = None
        self.a._rebuildgraph = True
        self.a._build_subgraph()

        source = self.p.lookup_type("src")
        target = self.p.lookup_type("tgt")
        flow_true = self.p.lookup_type("flow_true")
        flow_false = self.p.lookup_type("flow_false")

        r = self.a.G.edges[source, flow_true]["rules"]
        self.assertEqual(len(r), 1)
        r = self.a.G.edges[flow_true, target]["rules"]
        self.assertEqual(len(r), 1)
        r = self.a.G.edges[source, flow_false]["rules"]
        self.assertEqual(len(r), 1)
        r = self.a.G.edges[flow_false, target]["rules"]
        self.assertEqual(len(r), 1)

    def test_002_default_conditional_rules(self):
        """Keep only default conditional rules."""
        self.a.booleans = {}
        self.a._rebuildgraph = True
        self.a._build_subgraph()

        source = self.p.lookup_type("src")
        target = self.p.lookup_type("tgt")
        flow_true = self.p.lookup_type("flow_true")
        flow_false = self.p.lookup_type("flow_false")

        r = self.a.G.edges[source, flow_true]["rules"]
        self.assertEqual(len(r), 0)
        r = self.a.G.edges[flow_true, target]["rules"]
        self.assertEqual(len(r), 0)
        r = self.a.G.edges[source, flow_false]["rules"]
        self.assertEqual(len(r), 1)
        r = self.a.G.edges[flow_false, target]["rules"]
        self.assertEqual(len(r), 1)

    def test_003_user_conditional_true(self):
        """Keep only conditional rules selected by user specified booleans (True Case.)"""
        self.a.booleans = {"condition": True}
        self.a.rebuildgraph = True
        self.a._build_subgraph()

        source = self.p.lookup_type("src")
        target = self.p.lookup_type("tgt")
        flow_true = self.p.lookup_type("flow_true")
        flow_false = self.p.lookup_type("flow_false")

        r = self.a.G.edges[source, flow_true]["rules"]
        self.assertEqual(len(r), 1)
        r = self.a.G.edges[flow_true, target]["rules"]
        self.assertEqual(len(r), 1)
        r = self.a.G.edges[source, flow_false]["rules"]
        self.assertEqual(len(r), 0)
        r = self.a.G.edges[flow_false, target]["rules"]
        self.assertEqual(len(r), 0)

    def test_004_user_conditional_false(self):
        """Keep only conditional rules selected by user specified booleans (False Case.)"""
        self.a.booleans = {"condition": False}
        self.a.rebuildgraph = True
        self.a._build_subgraph()

        source = self.p.lookup_type("src")
        target = self.p.lookup_type("tgt")
        flow_true = self.p.lookup_type("flow_true")
        flow_false = self.p.lookup_type("flow_false")

        r = self.a.G.edges[source, flow_true]["rules"]
        self.assertEqual(len(r), 0)
        r = self.a.G.edges[flow_true, target]["rules"]
        self.assertEqual(len(r), 0)
        r = self.a.G.edges[source, flow_false]["rules"]
        self.assertEqual(len(r), 1)
        r = self.a.G.edges[flow_false, target]["rules"]
        self.assertEqual(len(r), 1)

    def test_005_remaining_edges(self):
        """Keep edges when rules are deleted, but there are still remaining rules on the edge."""
        self.a.booleans = {}
        self.a.rebuildgraph = True
        self.a._build_subgraph()

        source = self.p.lookup_type("src_remain")
        target = self.p.lookup_type("tgt_remain")
        flow = self.p.lookup_type("flow_remain")

        r = self.a.G.edges[source, flow]["rules"]
        self.assertEqual(len(r), 1)
        self.assertEqual(str(r[0]), 'allow src_remain flow_remain:infoflow hi_w;')
        r = self.a.G.edges[flow, target]["rules"]
        self.assertEqual(len(r), 1)
        self.assertEqual(str(r[0]), 'allow tgt_remain flow_remain:infoflow hi_r;')



# Copyright 2014-2015, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
import logging
from socket import IPPROTO_TCP, IPPROTO_UDP

from .mixins import MatchContext
from .query import PolicyQuery
from .policyrep import PortconRange, PortconProtocol
from .util import match_range


class PortconQuery(MatchContext, PolicyQuery):

    """
    Port context query.

    Parameter:
    policy          The policy to query.

    Keyword Parameters/Class attributes:
    protocol        The protocol to match (socket.IPPROTO_TCP for
                    TCP or socket.IPPROTO_UDP for UDP)

    ports           A 2-tuple of the port range to match. (Set both to
                    the same value for a single port)
    ports_subset    If true, the criteria will match if it is a subset
                    of the portcon's range.
    ports_overlap   If true, the criteria will match if it overlaps
                    any of the portcon's range.
    ports_superset  If true, the criteria will match if it is a superset
                    of the portcon's range.
    ports_proper    If true, use proper superset/subset operations.
                    No effect if not using set operations.

    user            The criteria to match the context's user.
    user_regex      If true, regular expression matching
                    will be used on the user.

    role            The criteria to match the context's role.
    role_regex      If true, regular expression matching
                    will be used on the role.

    type_           The criteria to match the context's type.
    type_regex      If true, regular expression matching
                    will be used on the type.

    range_          The criteria to match the context's range.
    range_subset    If true, the criteria will match if it is a subset
                    of the context's range.
    range_overlap   If true, the criteria will match if it overlaps
                    any of the context's range.
    range_superset  If true, the criteria will match if it is a superset
                    of the context's range.
    range_proper    If true, use proper superset/subset operations.
                    No effect if not using set operations.
    """

    _protocol = None
    _ports = None
    ports_subset = False
    ports_overlap = False
    ports_superset = False
    ports_proper = False

    @property
    def ports(self):
        return self._ports

    @ports.setter
    def ports(self, value):
        pending_ports = PortconRange(*value)

        if all(pending_ports):
            if pending_ports.low < 1 or pending_ports.high < 1:
                raise ValueError("Port numbers must be positive: {0.low}-{0.high}".
                                 format(pending_ports))

            if pending_ports.low > pending_ports.high:
                raise ValueError(
                    "The low port must be smaller than the high port: {0.low}-{0.high}".
                    format(pending_ports))

            self._ports = pending_ports
        else:
            self._ports = None

    @property
    def protocol(self):
        return self._protocol

    @protocol.setter
    def protocol(self, value):
        if value:
            self._protocol = PortconProtocol.lookup(value)
        else:
            self._protocol = None

    def __init__(self, policy, **kwargs):
        super(PortconQuery, self).__init__(policy, **kwargs)
        self.log = logging.getLogger(__name__)

    def results(self):
        """Generator which yields all matching portcons."""
        self.log.info("Generating portcon results from {0.policy}".format(self))
        self.log.debug("Ports: {0.ports}, overlap: {0.ports_overlap}, "
                       "subset: {0.ports_subset}, superset: {0.ports_superset}, "
                       "proper: {0.ports_proper}".format(self))
        self.log.debug("Protocol: {0.protocol!r}".format(self))
        self._match_context_debug(self.log)

        for portcon in self.policy.portcons():

            if self.ports and not match_range(
                    portcon.ports,
                    self.ports,
                    self.ports_subset,
                    self.ports_overlap,
                    self.ports_superset,
                    self.ports_proper):
                continue

            if self.protocol and self.protocol != portcon.protocol:
                continue

            if not self._match_context(portcon.context):
                continue

            yield portcon



# Copyright 2018, Chris PeBenito <pebenito@ieee.org>
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SETools.  If not, see <http://www.gnu.org/licenses/>.
#
import os
import unittest

from setools import IbpkeyconQuery

from .policyrep.util import compile_policy


class IbpkeyconQueryTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.p = compile_policy("tests/ibpkeyconquery.conf")

    @classmethod
    def tearDownClass(cls):
        os.unlink(cls.p.path)

    def test_000_unset(self):
        """ibpkeycon query with no criteria"""
        # query with no parameters gets all ibpkeycons.
        ibpkeycons = sorted(self.p.ibpkeycons())

        q = IbpkeyconQuery(self.p)
        q_ibpkeycons = sorted(q.results())

        self.assertListEqual(ibpkeycons, q_ibpkeycons)

    def test_001_subnet_mask(self):
        """Ibpkeycon query with subnet mask match."""
        q = IbpkeyconQuery(self.p, subnet_prefix="fe81::")

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(1, 1)], ibpkeycons)

    def test_010_pkey_exact(self):
        """Ibpkeycon query with exact pkey match."""
        q = IbpkeyconQuery(self.p, pkeys=(0x10c, 0x10e))

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(0x10c, 0x10e)], ibpkeycons)

    def test_020_user_exact(self):
        """ibpkeycon query with context user exact match"""
        q = IbpkeyconQuery(self.p, user="user20", user_regex=False)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(20, 20)], ibpkeycons)

    def test_021_user_regex(self):
        """ibpkeycon query with context user regex match"""
        q = IbpkeyconQuery(self.p, user="user21(a|b)", user_regex=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(0x21a, 0x21a), (0x21b, 0x21b)], ibpkeycons)

    def test_030_role_exact(self):
        """ibpkeycon query with context role exact match"""
        q = IbpkeyconQuery(self.p, role="role30_r", role_regex=False)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(30, 30)], ibpkeycons)

    def test_031_role_regex(self):
        """ibpkeycon query with context role regex match"""
        q = IbpkeyconQuery(self.p, role="role31(a|c)_r", role_regex=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(0x31a, 0x31a), (0x31c, 0x31c)], ibpkeycons)

    def test_040_type_exact(self):
        """ibpkeycon query with context type exact match"""
        q = IbpkeyconQuery(self.p, type_="type40", type_regex=False)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(40, 40)], ibpkeycons)

    def test_041_type_regex(self):
        """ibpkeycon query with context type regex match"""
        q = IbpkeyconQuery(self.p, type_="type41(b|c)", type_regex=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(0x41b, 0x41b), (0x41c, 0x41c)], ibpkeycons)

    def test_050_range_exact(self):
        """ibpkeycon query with context range exact match"""
        q = IbpkeyconQuery(self.p, range_="s0:c1 - s0:c0.c4")

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(50, 50)], ibpkeycons)

    def test_051_range_overlap1(self):
        """ibpkeycon query with context range overlap match (equal)"""
        q = IbpkeyconQuery(self.p, range_="s1:c1 - s1:c0.c4", range_overlap=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(51, 51)], ibpkeycons)

    def test_051_range_overlap2(self):
        """ibpkeycon query with context range overlap match (subset)"""
        q = IbpkeyconQuery(self.p, range_="s1:c1,c2 - s1:c0.c3", range_overlap=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(51, 51)], ibpkeycons)

    def test_051_range_overlap3(self):
        """ibpkeycon query with context range overlap match (superset)"""
        q = IbpkeyconQuery(self.p, range_="s1 - s1:c0.c4", range_overlap=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(51, 51)], ibpkeycons)

    def test_051_range_overlap4(self):
        """ibpkeycon query with context range overlap match (overlap low level)"""
        q = IbpkeyconQuery(self.p, range_="s1 - s1:c1,c2", range_overlap=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(51, 51)], ibpkeycons)

    def test_051_range_overlap5(self):
        """ibpkeycon query with context range overlap match (overlap high level)"""
        q = IbpkeyconQuery(self.p, range_="s1:c1,c2 - s1:c0.c4", range_overlap=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(51, 51)], ibpkeycons)

    def test_052_range_subset1(self):
        """ibpkeycon query with context range subset match"""
        q = IbpkeyconQuery(self.p, range_="s2:c1,c2 - s2:c0.c3", range_overlap=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(52, 52)], ibpkeycons)

    def test_052_range_subset2(self):
        """ibpkeycon query with context range subset match (equal)"""
        q = IbpkeyconQuery(self.p, range_="s2:c1 - s2:c1.c3", range_overlap=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(52, 52)], ibpkeycons)

    def test_053_range_superset1(self):
        """ibpkeycon query with context range superset match"""
        q = IbpkeyconQuery(self.p, range_="s3 - s3:c0.c4", range_superset=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(53, 53)], ibpkeycons)

    def test_053_range_superset2(self):
        """ibpkeycon query with context range superset match (equal)"""
        q = IbpkeyconQuery(self.p, range_="s3:c1 - s3:c1.c3", range_superset=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(53, 53)], ibpkeycons)

    def test_054_range_proper_subset1(self):
        """ibpkeycon query with context range proper subset match"""
        q = IbpkeyconQuery(self.p, range_="s4:c1,c2", range_subset=True, range_proper=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(54, 54)], ibpkeycons)

    def test_054_range_proper_subset2(self):
        """ibpkeycon query with context range proper subset match (equal)"""
        q = IbpkeyconQuery(self.p, range_="s4:c1 - s4:c1.c3", range_subset=True, range_proper=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([], ibpkeycons)

    def test_054_range_proper_subset3(self):
        """ibpkeycon query with context range proper subset match (equal low only)"""
        q = IbpkeyconQuery(self.p, range_="s4:c1 - s4:c1.c2", range_subset=True, range_proper=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(54, 54)], ibpkeycons)

    def test_054_range_proper_subset4(self):
        """ibpkeycon query with context range proper subset match (equal high only)"""
        q = IbpkeyconQuery(self.p, range_="s4:c1,c2 - s4:c1.c3", range_subset=True,
                           range_proper=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(54, 54)], ibpkeycons)

    def test_055_range_proper_superset1(self):
        """ibpkeycon query with context range proper superset match"""
        q = IbpkeyconQuery(self.p, range_="s5 - s5:c0.c4", range_superset=True, range_proper=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(55, 55)], ibpkeycons)

    def test_055_range_proper_superset2(self):
        """ibpkeycon query with context range proper superset match (equal)"""
        q = IbpkeyconQuery(self.p, range_="s5:c1 - s5:c1.c3", range_superset=True,
                           range_proper=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([], ibpkeycons)

    def test_055_range_proper_superset3(self):
        """ibpkeycon query with context range proper superset match (equal low)"""
        q = IbpkeyconQuery(self.p, range_="s5:c1 - s5:c1.c4", range_superset=True,
                           range_proper=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(55, 55)], ibpkeycons)

    def test_055_range_proper_superset4(self):
        """ibpkeycon query with context range proper superset match (equal high)"""
        q = IbpkeyconQuery(self.p, range_="s5 - s5:c1.c3", range_superset=True,
                           range_proper=True)

        ibpkeycons = sorted(n.pkeys for n in q.results())
        self.assertListEqual([(55, 55)], ibpkeycons)

    def test_900_invalid_subnet_prefix(self):
        """Ibpkeycon query with invalid subnet prefix"""
        with self.assertRaises(ValueError):
            IbpkeyconQuery(self.p, subnet_prefix="INVALID")

    def test_910_invalid_pkey_negative(self):
        """Ibpkeycon query with negative pkey"""
        with self.assertRaises(ValueError):
            IbpkeyconQuery(self.p, pkeys=(-1, -1))

        with self.assertRaises(ValueError):
            IbpkeyconQuery(self.p, pkeys=(1, -1))

        with self.assertRaises(ValueError):
            IbpkeyconQuery(self.p, pkeys=(-1, 1))

    def test_911_invalid_pkey_zero(self):
        """Ibpkeycon query with 0 pkey"""
        with self.assertRaises(ValueError):
            IbpkeyconQuery(self.p, pkeys=(0, 0))

    def test_912_invalid_pkey_over_max(self):
        """Ibpkeycon query with pkey over maximum value"""
        with self.assertRaises(ValueError):
            IbpkeyconQuery(self.p, pkeys=(1, 0xfffff))

        with self.assertRaises(ValueError):
            IbpkeyconQuery(self.p, pkeys=(0xfffff, 1))

        with self.assertRaises(ValueError):
            IbpkeyconQuery(self.p, pkeys=(0xfffff, 0xfffff))

    def test_913_invalid_pkey_not_a_number(self):
        """Ibpkeycon query with pkey is not a number"""
        with self.assertRaises(TypeError):
            IbpkeyconQuery(self.p, pkeys=(1, "INVALID"))

        with self.assertRaises(TypeError):
            IbpkeyconQuery(self.p, pkeys=("INVALID", 2))

    def test_914_invalid_pkey_not_tuple(self):
        """Ibpkeycon query with pkey is not a tuple"""
        with self.assertRaises(TypeError):
            IbpkeyconQuery(self.p, pkeys=1)

    def test_915_invalid_pkey_wrong_tuple_length(self):
        """Ibpkeycon query with pkey is not correct tuple size"""
        with self.assertRaises(TypeError):
            IbpkeyconQuery(self.p, pkeys=(1,))

        with self.assertRaises(TypeError):
            IbpkeyconQuery(self.p, pkeys=(1, 2, 3))



"""The SETools SELinux policy analysis library."""
# Copyright 2014-2015, Tresys Technology, LLC
# Copyright 2016, Chris PeBenito <pebenito@ieee.org>
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
try:
    import pkg_resources
    # pylint: disable=no-member
    __version__ = pkg_resources.get_distribution("setools").version
except ImportError:  # pragma: no cover
    __version__ = "unknown"

import logging

# Python classes for policy representation
from .policyrep import SELinuxPolicy, BoundsRuletype, ConstraintRuletype, DefaultRuletype, \
    DefaultRangeValue, DefaultValue, FSUseRuletype, HandleUnknown, IbpkeyconRange, MLSRuletype, \
    NodeconIPVersion, PolicyTarget, PortconProtocol, RBACRuletype, TERuletype

# Exceptions
from . import exception

# Component Queries
from .boolquery import BoolQuery
from .categoryquery import CategoryQuery
from .commonquery import CommonQuery
from .objclassquery import ObjClassQuery
from .polcapquery import PolCapQuery
from .rolequery import RoleQuery
from .sensitivityquery import SensitivityQuery
from .typequery import TypeQuery
from .typeattrquery import TypeAttributeQuery
from .userquery import UserQuery

# Rule Queries
from .mlsrulequery import MLSRuleQuery
from .rbacrulequery import RBACRuleQuery
from .terulequery import TERuleQuery

# Constraint queries
from .constraintquery import ConstraintQuery

# Other queries
from .boundsquery import BoundsQuery
from .defaultquery import DefaultQuery

# In-policy Context Queries
from .fsusequery import FSUseQuery
from .genfsconquery import GenfsconQuery
from .ibendportconquery import IbendportconQuery
from .ibpkeyconquery import IbpkeyconQuery
from .initsidquery import InitialSIDQuery
from .netifconquery import NetifconQuery
from .nodeconquery import NodeconQuery
from .portconquery import PortconQuery
from .ioportconquery import IoportconQuery
from .iomemconquery import IomemconQuery
from .pirqconquery import PirqconQuery
from .pcideviceconquery import PcideviceconQuery
from .devicetreeconquery import DevicetreeconQuery

# Information Flow Analysis
from .infoflow import InfoFlowAnalysis
from .permmap import PermissionMap

# Domain Transition Analysis
from .dta import DomainTransitionAnalysis

# Policy difference
from .diff import PolicyDifference

logging.getLogger(__name__).addHandler(logging.NullHandler())



# Copyright 2018, Chris PeBenito <pebenito@ieee.org>
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
from collections import namedtuple

from .context import ContextWrapper
from .descriptors import DiffResultDescriptor
from .difference import Difference, Wrapper


modified_ibendportcon_record = namedtuple("modified_ibendportcon", ["rule",
                                                                    "added_context",
                                                                    "removed_context"])


class IbendportconsDifference(Difference):

    """Determine the difference in ibendportcons between two policies."""

    added_ibendportcons = DiffResultDescriptor("diff_ibendportcons")
    removed_ibendportcons = DiffResultDescriptor("diff_ibendportcons")
    modified_ibendportcons = DiffResultDescriptor("diff_ibendportcons")

    def diff_ibendportcons(self):
        """Generate the difference in ibendportcons between the policies."""

        self.log.info(
            "Generating ibendportcon differences from {0.left_policy} to {0.right_policy}".
            format(self))

        self.added_ibendportcons, self.removed_ibendportcons, matched_ibendportcons = \
            self._set_diff(
                (IbendportconWrapper(n) for n in self.left_policy.ibendportcons()),
                (IbendportconWrapper(n) for n in self.right_policy.ibendportcons()))

        self.modified_ibendportcons = []

        for left_ibep, right_ibep in matched_ibendportcons:
            # Criteria for modified ibendportcons
            # 1. change to context
            if ContextWrapper(left_ibep.context) != ContextWrapper(right_ibep.context):
                self.modified_ibendportcons.append(
                    modified_ibendportcon_record(left_ibep,
                                                 right_ibep.context,
                                                 left_ibep.context))

    #
    # Internal functions
    #
    def _reset_diff(self):
        """Reset diff results on policy changes."""
        self.log.debug("Resetting ibendportcon differences")
        self.added_ibendportcons = None
        self.removed_ibendportcons = None
        self.modified_ibendportcons = None


class IbendportconWrapper(Wrapper):

    """Wrap ibendportcon statements for diff purposes."""

    __slots__ = ("name", "port")

    def __init__(self, ocon):
        self.origin = ocon
        self.name = ocon.name
        self.port = ocon.port
        self.key = hash(ocon)

    def __hash__(self):
        return self.key

    def __lt__(self, other):
        return self.origin < other.origin

    def __eq__(self, other):
        return self.name == other.name and \
            self.port == other.port



# Copyright 2014-2015, Tresys Technology, LLC
# Copyright 2018, Chris PeBenito <pebenito@ieee.org>
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#
from abc import ABC, abstractmethod


class PolicyQuery(ABC):

    """Abstract base class for SELinux policy queries."""

    def __init__(self, policy, **kwargs):
        self.policy = policy

        # keys are sorted in reverse order so regex settings
        # are set before the criteria, e.g. name_regex
        # is set before name.  This ensures correct behavior
        # since the criteria descriptors are sensitve to
        # regex settings.
        for name in sorted(kwargs.keys(), reverse=True):
            attr = getattr(self, name, None)  # None is not callable
            if callable(attr):
                raise ValueError("Keyword parameter {0} conflicts with a callable.".format(name))

            setattr(self, name, kwargs[name])

    @abstractmethod
    def results(self):
        """
        Generator which returns the matches for the query.  This method
        should be overridden by subclasses.
        """
        pass



# Copyright 2016, Tresys Technology, LLC
#
# This file is part of SETools.
#
# SETools is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1 of
# the License, or (at your option) any later version.
#
# SETools is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with SETools.  If not, see
# <http://www.gnu.org/licenses/>.
#


def match_regex(obj, criteria, regex):
    """
    Match the object with optional regular expression.

    Parameters:
    obj         The object to match.
    criteria    The criteria to match.
    regex       If regular expression matching should be used.
    """

    if regex:
        return bool(criteria.search(str(obj)))
    else:
        return obj == criteria


def match_set(obj, criteria, equal):
    """
    Match the object (a set) with optional set equality.

    Parameters:
    obj         The object to match. (a set)
    criteria    The criteria to match. (a set)
    equal       If set equality should be used. Otherwise
                any set intersection will match.
    """

    if equal:
        return obj == criteria
    else:
        return bool(obj.intersection(criteria))


def match_in_set(obj, criteria, regex):
    """
    Match if the criteria is in the list, with optional
    regular expression matching.

    Parameters:
    obj         The object to match.
    criteria    The criteria to match.
    regex       If regular expression matching should be used.
    """

    if regex:
        return [m for m in obj if criteria.search(str(m))]
    else:
        return criteria in obj


def match_indirect_regex(obj, criteria, indirect, regex):
    """
    Match the object with optional regular expression and indirection.

    Parameters:
    obj         The object to match.
    criteria    The criteria to match.
    regex       If regular expression matching should be used.
    indirect    If object indirection should be used, e.g.
                expanding an attribute.
    """

    if indirect:
        if regex:
            return [o for o in obj.expand() if criteria.search(str(o))]
        else:
            return set(criteria.expand()).intersection(obj.expand())
    else:
        return match_regex(obj, criteria, regex)


def match_regex_or_set(obj, criteria, equal, regex):
    """
    Match the object (a set) with either set comparisons
    (equality or intersection) or by regex matching of the
    set members.  Regular expression matching will override
    the set equality option.

    Parameters:
    obj         The object to match. (a set)
    criteria    The criteria to match.
    equal       If set equality should be used.  Otherwise
                any set intersection will match. Ignored
                if regular expression matching is used.
    regex       If regular expression matching should be used.
    """

    if regex:
        return [m for m in obj if criteria.search(str(m))]
    else:
        return match_set(obj, set(criteria), equal)


def match_range(obj, criteria, subset, overlap, superset, proper):
    """
    Match ranges of objects.

    obj         An object with attributes named "low" and "high", representing the range.
    criteria    An object with attributes named "low" and "high", representing the criteria.
    subset      If true, the criteria will match if it is a subset obj's range.
    overlap     If true, the criteria will match if it overlaps any of the obj's range.
    superset    If true, the criteria will match if it is a superset of the obj's range.
    proper      If true, use proper superset/subset operations.
                No effect if not using set operations.
    """

    if overlap:
        return ((obj.low <= criteria.low <= obj.high) or (
            obj.low <= criteria.high <= obj.high) or (
            criteria.low <= obj.low and obj.high <= criteria.high))
    elif subset:
        if proper:
            return ((obj.low < criteria.low and criteria.high <= obj.high) or (
                obj.low <= criteria.low and criteria.high < obj.high))
        else:
            return obj.low <= criteria.low and criteria.high <= obj.high
    elif superset:
        if proper:
            return ((criteria.low < obj.low and obj.high <= criteria.high) or (
                criteria.low <= obj.low and obj.high < criteria.high))
        else:
            return (criteria.low <= obj.low and obj.high <= criteria.high)
    else:
        return criteria.low == obj.low and obj.high == criteria.high


def match_level(obj, criteria, dom, domby, incomp):
    """
    Match the an MLS level.

    obj         The level to match.
    criteria    The criteria to match. (a level)
    dom         If true, the criteria will match if it dominates obj.
    domby       If true, the criteria will match if it is dominated by obj.
    incomp      If true, the criteria will match if it is incomparable to obj.
    """

    if dom:
        return (criteria >= obj)
    elif domby:
        return (criteria <= obj)
    elif incomp:
        return (criteria ^ obj)
    else:
        return (criteria == obj)
